/*
 * Derived from linux 2.4.28
 * IPSec Passthrough Modifier: Chad Sheu
 */ 

//#include <linux/config.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/timer.h>
#include <linux/netfilter.h>
#include <linux/ip.h>
#include <linux/in.h>
#include <linux/list.h>
#include <linux/seq_file.h>
#include <linux/interrupt.h>
#include <linux/delay.h>

static DEFINE_RWLOCK(esp_lock);
#define ASSERT_READ_LOCK(x) 
#define ASSERT_WRITE_LOCK(x)

#include <linux/netfilter_ipv4/ip_conntrack_protocol.h>
#include <linux/netfilter_ipv4/ip_conntrack_helper.h>
#include <linux/netfilter_ipv4/ip_conntrack_core.h>
#include <linux/netfilter_ipv4/ip_conntrack_proto_esp.h>

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Chad Sheu");
MODULE_DESCRIPTION("Netfilter connection tracking protocol helper for ESP");

/* defined from kernel 2.4 version, modify if necessary */
#define ESP_TIMEOUT (100*HZ)
#define ESP_STREAM_TIMEOUT (600*HZ)	

#if 0
    #define DEBUGP(format, args...)  printk(KERN_INFO  __FILE__ ":" __FUNCTION__ ":" format ,  ## args)
    #define DUMP_TUPLE_ESP(x,str) printk(" [%s] SRC:(%u.%u.%u.%u):SPI(0x%x) --->  DST (%u.%u.%u.%u):SPI(0x%x) lastspi(0x%x)\n", \
		str,NIPQUAD((x)->sip), ntohl((x)->spi_local), \
		NIPQUAD((x)->dip), ntohl((x)->spi_remote),ntohl(last_spi))
#else
    #define DEBUGP(x, args...)
    #define DUMP_TUPLE_ESP(x,str)
#endif

/* ESP SPI mapping Handling Function*/
#define LIST_FIND(head, cmpfn, type, args...)       \
({                                                  \
     const struct list_head *__i, *__j = NULL;      \
	                                                \
     ASSERT_READ_LOCK(head);                        \
     if ((head)->next == NULL)                      \
        last_spi=0;                                 \
                                                    \
     list_for_each(__i, (head))                     \
         if (cmpfn((const type)__i , ## args)) {    \
             __j = __i;                             \
             break;                                 \
         }                                          \
     (type)__j;                                     \
})

#define LIST_LIST(head, cmpfn, type)                \
({                                                  \
     const struct list_head *__i;                   \
                                                    \
     ASSERT_READ_LOCK(head);                        \
     printk("\n----- List Map -----\n");            \
     list_for_each(__i, (head))                     \
         cmpfn((const type)__i);                    \
     printk("----------------------\n");            \
})

static LIST_HEAD(esp_spi_list);
static u_int32_t dir =0, last_spi =0;
static u_int16_t re_index = 0;

#if 0
static u_int32_t lanip = 0,netmask = 0;
static u_int32_t _atol(unsigned char *buf)
{
	u_int16_t len = strlen(buf);
	u_int16_t i=0,j=0;
	unsigned long ret_value = 0,tmp_value=0;

    //printk("Length = %d\n",len);
	for (i=0;i<len;i++) {
		j=0;
        //printk("buf [%d] = %c\n",i,buf[i]);
		tmp_value = 0;
		tmp_value = *(buf+i) - '0';
        //printk("1 tmp_value = %d\n",tmp_value);
       
		while(j< (len-i-1)) {
			tmp_value *= 10;
			j++;
		}

        //printk("2 tmp_value = %d\n",tmp_value);
		ret_value += (unsigned int)tmp_value;
	}
    //printk("%s:WANIP = %d\n",__FUNCTION__,ret_value);
	return ret_value;
}

static u_int32_t _swap_int(u_int32_t org)
{
	return (((org & 0xFF000000) >>24) |
			((org & 0x00FF0000) >> 8) |
			((org & 0x0000FF00) << 8) |
			((org & 0x000000FF) << 24));
}

// out packet return 1
// in packet return 0
static u_int32_t esp_in_out_check(struct ip_conntrack_tuple *orig_tuple)
{
	struct file * fd = NULL;       //Laroche add
	if (lanip == 0){
		char __user buf[12];

		//Get NetMask
		fd = filp_open("/var/config/NM",O_RDONLY,0660);
		if (!IS_ERR(fd)) {
			mm_segment_t orig_fs;
			memset(buf,0,12);
			orig_fs = get_fs();
			set_fs(KERNEL_DS);
			do_sync_read(fd, buf, 12,&fd->f_pos);
			set_fs(orig_fs);
			filp_close(fd, NULL);
			buf[strlen(buf)-1] = '\0';
		}
		else {
            //printk("netmask not get 1\n");
			netmask = 0x00ffffff;
			return 0;
		}
		
		if (strlen(buf) > 2) {	
			netmask = _atol(buf);
		}
		else {
            //printk("netmask not get 2\n");
			netmask = 0x00ffffff;
			return 1;
		}

		fd = filp_open("/var/config/LAN_ip",O_RDONLY,0660);
		if (!IS_ERR(fd)) {
			mm_segment_t orig_fs;
			memset(buf,0,12);
			orig_fs = get_fs();
			set_fs(KERNEL_DS);
			do_sync_read(fd, buf, 12,&fd->f_pos);
			set_fs(orig_fs);
			filp_close(fd, NULL);
			buf[strlen(buf)-1] = '\0';
		}
		else {
            //printk("lan not get 1\n");
			lanip = 0x00000000;
			return 0;
		}
		
		if (strlen(buf) > 2) {	
			lanip = _atol(buf);
		}
		else {
            //printk("wan not get 2\n");
			lanip = 0x00000000;
			return 1;
		}

		if (lanip != 0 ) 
            lanip &= netmask;
	}
    //printk("LANN IP [%u.%u.%u.%u] DEST IP [%u.%u.%u.%u]\n",
    //			NIPQUAD(lanip),NIPQUAD(orig_tuple->dst.ip));
	
	if (lanip == 0)
		return 0;
	if ((orig_tuple->src.ip & netmask ) == lanip) {
        //printk("is Out packet\n");
		return 1;
	}
	else {
        //printk("is In packet\n");
		return 0;
	}
}
#endif

static void __del_esp(struct ip_ct_esp_spi *trig)
{
	DUMP_TUPLE_ESP(&trig->esp_master,"esp_del");
	
    /* delete from 'trigger_list' */
    write_lock_bh(&esp_lock);	 

	list_del(&trig->list);
	kfree(trig);

    write_unlock_bh(&esp_lock);	
}

void esp_timeout(unsigned long trig)
{
	struct ip_ct_esp_spi *esp_trig = (struct ip_ct_esp_spi *)trig;
	last_spi = 0 ;
    DUMP_TUPLE_ESP(&esp_trig->esp_master,"esp_timeout");
   
    write_lock_bh(&esp_lock);
    
	__del_esp(esp_trig);
    
	write_unlock_bh(&esp_lock);
}

static void esp_refresh(struct ip_ct_esp_spi *trig, unsigned long extra_jiffies)
{
	//DEBUGP("\n@@@@ %s @@@@\n", __FUNCTION__);
    IP_NF_ASSERT(trig);

	write_lock_bh(&esp_lock);

    /* Need del_timer for race avoidance (may already be dying). */
	if (del_timer(&trig->timeout)) {
		trig->timeout.expires = jiffies + extra_jiffies;
		add_timer(&trig->timeout);
	}

    write_unlock_bh(&esp_lock);
}

int esp_packet_list(const struct ip_ct_esp_spi* esp)
{
	printk("Source IP [%u.%u.%u.%u] DEST IP [%u.%u.%u.%u]\n",
			NIPQUAD(esp->esp_master.sip),NIPQUAD(esp->esp_master.dip));
	printk("spi_local = [0x%x] spi_remote = [0x%x]\n\n",
			ntohl(esp->esp_master.spi_local),ntohl(esp->esp_master.spi_remote));

	return 1;
}

int esp_packet_matched(const struct ip_ct_esp_spi* esp,const u_int32_t now_spi,
                    const u_int32_t saddr,const u_int32_t daddr, u_int32_t *dir)
{
	u_int32_t spi = 0;
	u_int32_t ret = 0;

	if( (esp->esp_master.spi_local == now_spi)
	    && (esp->esp_master.sip == saddr) && (esp->esp_master.dip==daddr)) 
    {
		//Xhʥ]t
		*dir = IP_CT_DIR_ORIGINAL;
		DUMP_TUPLE_ESP(&esp->esp_master," match source");
		ret = 1;
	}
	else if((esp->esp_master.spi_local == now_spi) && (esp->esp_master.spi_remote == 0) && 
			(esp->esp_master.dip==saddr)) 
    {
		//Ĥ@Xhʥ] ^Ӫɭԭnt
		*dir = IP_CT_DIR_REPLY;
		DUMP_TUPLE_ESP(&esp->esp_master,"first match destination");
		ret = 1;
	}
	else if((esp->esp_master.spi_remote == now_spi) && 
		   (esp->esp_master.dip==saddr) )
	{
		// VPN SERVER --> Routerʥ]
		*dir = IP_CT_DIR_REPLY;
		DUMP_TUPLE_ESP(&esp->esp_master,"match destination");
		ret = 1;
	}
	else {
		//map䤣
		DUMP_TUPLE_ESP(&esp->esp_master,"no match");
        //printk("SRC:(%u.%u.%u.%u), DST:(%u.%u.%u.%u), SPI:0x%x\n",NIPQUAD(saddr),NIPQUAD(daddr),ntohl(now_spi));
	}

	return ret;
}

int ip_spi_add(struct ip_ct_esp_master *new_tuple)
{
	struct ip_ct_esp_spi* esp_new = NULL;

	//add new record for src_spi
	esp_new = kmalloc(sizeof(*esp_new), GFP_ATOMIC);
	if (!esp_new) {
		printk("ESP malloc memory is error.\n");
		return -1;
	}
	
	write_lock_bh(&esp_lock);

	/* initializing list head should be sufficient memory */	
	memset(esp_new, 0, sizeof(*esp_new));
	memcpy(&(esp_new->esp_master), new_tuple, sizeof(struct ip_ct_esp_master));
    
	/* add and start timer if required */
	init_timer(&esp_new->timeout);
    esp_new->timeout.data = (unsigned long)esp_new;
    esp_new->timeout.function = &esp_timeout;
    esp_new->timeout.expires = jiffies + (ESP_TIMEOUT);
	add_timer(&esp_new->timeout);
	
	//list_append(&esp_spi_list, esp_new);
	list_add_tail(&esp_new->list,&esp_spi_list);
    	
	write_unlock_bh(&esp_lock);
	
	DUMP_TUPLE_ESP(&esp_new->esp_master,"add tuple");
	return NF_ACCEPT;
}

static u_int32_t esp_spi_lookup(struct ip_conntrack_tuple *orig_tuple,const u_int32_t now_spi)
{
	struct ip_ct_esp_spi *found = NULL, *found2 = NULL;
	struct ip_conntrack_tuple tuple;
	u_int32_t dir;
	
    //printk("\n----- %s:%d %s() -----\n",__FILE__,__LINE__,__FUNCTION__);
    //printk("1. SRC IP %u.%u.%u.%u -> DST IP %u.%u.%u.%u \n",NIPQUAD(orig_tuple->src.ip),NIPQUAD(orig_tuple->dst.ip));
    //LIST_LIST(&esp_spi_list, esp_packet_list,struct ip_ct_esp_spi *);
    //printk("SPI = 	[0x%x ] \n",ntohl(now_spi));
	
	read_lock_bh(&esp_lock);
	memset(&tuple, 0, sizeof(tuple));
	memcpy(&tuple,orig_tuple,sizeof(struct ip_conntrack_tuple));
	
	found = LIST_FIND(&esp_spi_list, esp_packet_matched,
			struct ip_ct_esp_spi *,now_spi, orig_tuple->src.ip, orig_tuple->dst.ip,&dir);

	if (!found){  //not found in original
        //printk("Not found in original \n");
        //printk("Last SPI = 0x%x\n",ntohl(last_spi));
		//if ((esp_in_out_check(orig_tuple))){ // just out packet to add a list
			if (last_spi == 0) {
				struct ip_ct_esp_master esp_master;
				esp_master.sip = orig_tuple->src.ip;
				esp_master.dip = orig_tuple->dst.ip;
				esp_master.spi_local = now_spi;
				esp_master.spi_remote = 0;
				ip_spi_add(&esp_master);
			
				orig_tuple->src.u.esp.spi = now_spi;
				orig_tuple->dst.u.esp.spi = 0;	
			
                //printk(KERN_INFO "Not found !! Is the new src_spi\n");
				last_spi = now_spi;
				goto end;
			}
		//}
	    found2 = LIST_FIND(&esp_spi_list, esp_packet_matched,
				struct ip_ct_esp_spi *, last_spi, orig_tuple->src.ip,orig_tuple->dst.ip,&dir);
	
		if (!found2) {
			re_index ++;
            //printk("Not Found 2 re_index = %d\n",re_index);
			if (re_index > 3) {
				last_spi = 0;
				re_index = 0;
			}
			goto fail;	
		}
		else {
			found2->esp_master.spi_remote = now_spi;
			orig_tuple->dst.u.esp.spi = found2->esp_master.spi_local;
			orig_tuple->src.u.esp.spi = 0;
			
			esp_refresh(found2, ESP_TIMEOUT);
			last_spi = 0;
			goto end;
		}
	}
	else{ //found != NULL
		if (dir == IP_CT_DIR_REPLY) {
			orig_tuple->dst.u.esp.spi = found->esp_master.spi_local;
			orig_tuple->src.u.esp.spi = 0;
            orig_tuple->dst.ip = found->esp_master.sip;
		}
		else{
		}
		
		esp_refresh(found, ESP_TIMEOUT);	
		//printk(KERN_INFO "found!!...\n" " sip[%u.%u.%u.%u] dip[%u.%u.%u.%u]\n", 
		//			NIPQUAD(found->esp_master.sip), NIPQUAD(found->esp_master.dip));
		goto end;
	}

fail:
end:
    read_unlock_bh(&esp_lock);
    //printk("2. SRC IP %u.%u.%u.%u -> DST IP %u.%u.%u.%u \n",NIPQUAD(orig_tuple->src.ip),NIPQUAD(orig_tuple->dst.ip));
    //printk("----------------------\n",__FILE__,__LINE__,__FUNCTION__);
    return NF_ACCEPT;
}

/*================================================================*/
/* function assigned to the field of protocol helper structure    */
/*================================================================*/
static int esp_pkt_to_tuple(const struct sk_buff *skb, unsigned dataoff,
                         struct ip_conntrack_tuple *tuple)
{
	struct esp_hdr hdr,*phdr;
	phdr = skb_header_pointer(skb, dataoff, sizeof(hdr), &hdr);
	const u_int32_t *spi = (u_int32_t *)(&phdr->spi);
	
    //printk("Function %s\n",__FUNCTION__);
    //printk("[SRC IP]:%u.%u.%u.%u -->[DST IP]:%u.%u.%u.%u\n",NIPQUAD(skb->nh.iph->saddr),NIPQUAD(skb->nh.iph->daddr));
    //printk("[PROTO NUM]: %d\n",skb->nh.iph->protocol);
    //printk("[SPI] : 0x%x\n",ntohl(*spi));
    //LIST_LIST(&esp_spi_list, esp_packet_list,struct ip_ct_esp_spi *);
	
	esp_spi_lookup(tuple,*spi);
	return 1;
}

static int esp_invert_tuple(struct ip_conntrack_tuple *inverse_tuple,
                        const struct ip_conntrack_tuple *orig_tuple)
{
    inverse_tuple->src.u.esp.spi = orig_tuple->dst.u.esp.spi;
    inverse_tuple->dst.u.esp.spi = orig_tuple->src.u.esp.spi;
    return 1;
}

/* Print out the per-protocol part of the tuple. */
static unsigned int esp_print_tuple(struct seq_file *buffer,
					const struct ip_conntrack_tuple *tuple)
{
	return seq_printf(buffer, "src spi = 0x%08x ", ntohl(tuple->src.u.esp.spi));	
}

/* Print out the private part of the conntrack. */
static unsigned int esp_print_conntrack(struct seq_file *buffer,
					    const struct ip_conntrack *state)
{
    //return seq_printf(buffer,"timeout = %u , streamtimeout = %u",(state->proto.esp.timeout/HZ),
    //                  (state->proto.esp.stream_timeout/HZ)); 
	return 0;
}

/* Returns verdict for packet, or -1 for invalid. */
static int esp_established(struct ip_conntrack *conntrack,
		       struct sk_buff *skb,enum ip_conntrack_info conntrackinfo)
{
	/* If we've seen traffic both ways, this is some kind of ESP stream. Extend timeout. */
	if (conntrack->status & IPS_SEEN_REPLY) {
		ip_ct_refresh_acct(conntrack,conntrackinfo,skb,ESP_STREAM_TIMEOUT);

		/* Also, more likely to be important, and not a probe */
		set_bit(IPS_ASSURED_BIT, &conntrack->status);
	} 
    else {
		ip_ct_refresh_acct(conntrack,conntrackinfo,skb,ESP_TIMEOUT);
	}
	return NF_ACCEPT;
}

/* Called when a new connection for this protocol found. */
static int esp_new(struct ip_conntrack *ct, const struct sk_buff *skb)
{   
	ct->proto.esp.stream_timeout = ESP_STREAM_TIMEOUT;	
	ct->proto.esp.timeout = ESP_TIMEOUT;	
	return 1;
}

/* Called when a conntrack entry has already been removed from the hashes
and is about to be deleted from memory*/
static void esp_destroy(struct ip_conntrack *ct)
{
	/* user defined */
}
/*================================================================*/

/* protocol helper structure definition*/
static struct ip_conntrack_protocol esp  = { 
    .proto = 50, 
    .name = "esp",
    .pkt_to_tuple    = esp_pkt_to_tuple, 
    .invert_tuple    = esp_invert_tuple, 
    .print_tuple     = esp_print_tuple,
    .print_conntrack = esp_print_conntrack, 
    .packet          = esp_established, 
    .new             = esp_new,
    .destroy         = NULL,
    .me              = THIS_MODULE,
    .tuple_to_nfattr = NULL,
    .nfattr_to_tuple = NULL, 
};

/* ip_conntrack_proto_esp initialization */
static int __init init(void)
{
	int retcode;
	if ((retcode = ip_conntrack_protocol_register(&esp))) {
        printk(KERN_ERR "Unable to register conntrack protocol "
			        "helper for esp: %d\n",	retcode);
		return -EIO;
	}
	printk("Register conntrack protocol helper for ESP...\n");
	return 0;
}

static void __exit fini(void)
{
	struct list_head *pos, *n;
	printk(" fini entry free register.\n");

	/* delete all keymap entries */
	write_lock_bh(&esp_lock);
	
	list_for_each_safe(pos, n, &esp_spi_list) {
		
		printk("deleting spi %p at module unload time\n", pos);
		list_del(pos);
		kfree(pos);
	}

	write_unlock_bh(&esp_lock);
	ip_conntrack_protocol_unregister(&esp); 
}

module_init(init);
module_exit(fini);

